# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from typing import Callable, Iterator, Sequence, Tuple

import networkx as grph

from cell_tracking_BC.standard.issue import ISSUE_SEPARATOR
from cell_tracking_BC.type.compartment.cell import cell_t


TIME_POINT = "time_point"  # Leave it here (as opposed to making it a class variable) since it is used "everywhere"


class unstructured_track_t(grph.DiGraph):
    in_degree: Callable[[cell_t], int]
    out_degree: Callable[[cell_t], int]

    def RootCellWithTimePoint(self) -> Tuple[cell_t, int]:
        """"""
        output = tuple(
            _rcd for _rcd in self.nodes.data(TIME_POINT) if self.in_degree(_rcd[0]) == 0
        )

        if (n_roots := output.__len__()) != 1:
            raise ValueError(
                f"{n_roots}{ISSUE_SEPARATOR}Invalid number of root cells. Expected=1."
            )

        return output[0]

    def LeafCellsWithTimePoints(self) -> Tuple[Sequence[cell_t], Sequence[int]]:
        """"""
        # TODO: Contact the Networkx team about the following comment (or check code on github)
        #     /!\ It seems that networkx.DiGraph.nodes.data does not guarantee the node enumeration order. This could be
        #     inconvenient for reproducibility checks.
        # The following sorting is undefined for 2 cells with the same centroid in the same frame, but this should not
        # happen in practice.
        records = sorted(
            (
                _rcd
                for _rcd in self.nodes.data(TIME_POINT)
                if self.out_degree(_rcd[0]) == 0
            ),
            key=lambda _rcd: str(_rcd[1]) + str(_rcd[0].centroid),
        )
        leaves, time_points = zip(*records)

        return tuple(leaves), tuple(time_points)

    @property
    def labels(self) -> Sequence[int]:
        """"""
        return (-1,)  # Note: this is a tuple

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        time_points = grph.get_node_attributes(self, TIME_POINT)

        for edge in self.edges:
            time_point = time_points[edge[0]]
            has_leaf = self.out_degree(edge[1]) == 0
            yield time_point, *edge, has_leaf

    @property
    def multi_segments_iterator(self) -> Iterator[Tuple[int, Sequence[cell_t], bool]]:
        """"""
        time_points = grph.get_node_attributes(self, TIME_POINT)

        graphs = [grph.DiGraph(self)]
        while graphs.__len__() > 0:
            current = graphs.pop()

            roots = sorted(
                (
                    _rcd
                    for _rcd in current.nodes.data(TIME_POINT)
                    if current.in_degree(_rcd[0]) == 0
                ),
                key=lambda _rcd: str(_rcd[1]) + str(_rcd[0].centroid),
            )
            leaves = sorted(
                (
                    _rcd
                    for _rcd in current.nodes.data(TIME_POINT)
                    if current.out_degree(_rcd[0]) == 0
                ),
                key=lambda _rcd: str(_rcd[1]) + str(_rcd[0].centroid),
            )

            longest_multi = grph.shortest_path(
                current, source=roots[0][0], target=leaves[-1][0]
            )
            time_point = time_points[longest_multi[0]]
            has_leaf = current.out_degree(longest_multi[-1]) == 0
            yield time_point, longest_multi, has_leaf

            edges = tuple(zip(longest_multi[:-1], longest_multi[1:]))
            current.remove_edges_from(edges)
            for cells in grph.weakly_connected_components(current):
                if cells.__len__() > 1:
                    track_view = self.subgraph(cells)
                    # Copy or re-instantiation is necessary since the subgraph is a view
                    graphs.append(grph.DiGraph(track_view))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import tempfile as temp
from pathlib import Path as path_t
from typing import IO, Union

import xlsxwriter as xlsx
from xlsxwriter.workbook import Workbook as workbook_t

from cell_tracking_BC.in_out.file.sheet import (
    SheetNameFromLongName,
    SortAndWriteCSVLines,
)
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.type.analysis import analysis_t
from cell_tracking_BC.type.track.multiple.structured import tracks_t
from cell_tracking_BC.type.track.single.structured import structured_track_t


INVALID_TRACK_MARKER = "Invalid"
PRUNED_TRACK_MARKER = "Pruned"
VALID_TRACK_MARKER = "Valid"


def SaveTrackDetailsToXLSX(
    path: Union[str, path_t],
    analysis: analysis_t,
    /,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if path.is_dir():
        path /= "tracks.xlsx"

    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))

    csv_path = path.with_suffix(".csv")
    csv_accessor = open(csv_path, mode="w")

    tracks = analysis.tracks
    if (tracks is None) or (tracks.__len__() == 0):
        LOGGER.warning("Sequence with no valid tracks.")
    else:
        _SaveSiblings(tracks, workbook, csv_accessor)
        _SaveTrackStatus(tracks, workbook, csv_accessor)

    workbook.close()
    csv_accessor.close()


def _SaveSiblings(
    tracks: tracks_t,
    workbook: workbook_t,
    csv_accessor: IO,
    /,
) -> None:
    """"""
    sheet_name = SheetNameFromLongName("siblings")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Siblings\n")

    csv_lines = []
    for track in tracks.all_structured_iterator:
        for c_idx, (labels, label_type) in enumerate(
            zip(
                (set(track.labels), set(track.topologic_labels)),
                ("", "Topologic "),
            )
        ):
            for label in labels:
                other_labels = str(sorted(labels.difference((label,))))[1:-1]
                if other_labels.__len__() > 0:
                    worksheet.write_string(label - 1, c_idx, other_labels)
                    csv_lines.append(f"{label}, {label_type}Labels: {other_labels}")
                else:
                    if label_type.__len__() > 0:
                        message = f"Thread track in {label_type}mode"
                    else:
                        message = "Thread track"
                    worksheet.write_string(label - 1, c_idx, message)
                    csv_lines.append(f"{label}, {message}")

    worksheet.write_string(tracks.total_n_topologic_threads, 0, "END")
    SortAndWriteCSVLines(csv_lines, csv_accessor)


def _SaveTrackStatus(
    tracks: tracks_t,
    workbook: workbook_t,
    csv_accessor: IO,
    /,
) -> None:
    """"""
    sheet_name = SheetNameFromLongName("status")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Status\n")

    csv_lines = []
    invalids = (
        _rcd[0] for _rcd in tracks.invalids if isinstance(_rcd[0], structured_track_t)
    )
    for tracks_per_status, status in zip(
        (tracks, tracks.fully_pruned, invalids),
        (VALID_TRACK_MARKER, PRUNED_TRACK_MARKER, INVALID_TRACK_MARKER),
    ):
        for track in tracks_per_status:
            for label in track.topologic_labels:
                if (label in track.labels) or (status != VALID_TRACK_MARKER):
                    thread_status = status
                else:
                    thread_status = PRUNED_TRACK_MARKER
                worksheet.write_string(label - 1, 0, thread_status)
                csv_lines.append(f"{label}, {thread_status}")

    worksheet.write_string(tracks.total_n_topologic_threads, 0, "END")
    SortAndWriteCSVLines(csv_lines, csv_accessor)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import datetime as dttm
import sys as sstm
from pathlib import Path as path_t
from typing import Union

from cell_tracking_BC.in_out.text.logger import LOGGER


# The following lists are meant to be safe enough, not to serve as references
PATH_ILLEGAL_CHARACTERS_LIN = r"/"
PATH_ILLEGAL_CHARACTERS_OSX = r":"
PATH_ILLEGAL_CHARACTERS_WIN = r'|/\<>:?*"'

REPLACEMENT_CHARACTER = "_"
VERSION_SEPARATOR = "-"

PATH_ILLEGAL_CHARACTERS = "".join(
    set(
        PATH_ILLEGAL_CHARACTERS_LIN
        + PATH_ILLEGAL_CHARACTERS_OSX
        + PATH_ILLEGAL_CHARACTERS_WIN
    )
)
if (REPLACEMENT_CHARACTER in PATH_ILLEGAL_CHARACTERS) or (
    VERSION_SEPARATOR in PATH_ILLEGAL_CHARACTERS
):
    raise ValueError(
        f'The character "{REPLACEMENT_CHARACTER}" or "{VERSION_SEPARATOR}" is an illegal path character'
    )


def StorageFolderForMain(
    main: Union[str, path_t],
    folder_name: str,
    /,
    *,
    base_folder_name: str = "_replay",
    base_folder_postfix: str = "",
) -> path_t:
    """
    Call with main equal to __file__
    """
    if isinstance(main, str):
        main = path_t(main)

    output = main.parent / (base_folder_name + base_folder_postfix)
    output = ReplacePathIllegalCharacters(output)
    if not output.exists():
        # exist_ok: added for thread safety (could use only mkdir then...)
        output.mkdir(exist_ok=True)
    elif not output.is_dir():
        LOGGER.error(f"{output}: Not a suitable storage folder")
        sstm.exit(-1)

    output /= ReplacePathIllegalCharacters(folder_name)
    if not output.exists():
        # exist_ok: added for thread safety (could use only mkdir then...)
        output.mkdir(exist_ok=True)
    elif not output.is_dir():
        LOGGER.error(f"{output}: Not a suitable storage folder")
        sstm.exit(-1)

    output /= TimeStamp()
    if output.exists():
        LOGGER.error(f"{output}: Existing date-based storage folder; Exiting")
        sstm.exit(-1)
    output.mkdir()

    return output


def TimeStamp() -> str:
    """"""
    return (
        dttm.datetime.now()
        .isoformat(timespec="milliseconds")
        .replace(".", "-")
        .replace(":", "-")
    )


def ReplacePathIllegalCharacters(
    path: Union[str, path_t], /, *, replacement: str = REPLACEMENT_CHARACTER
) -> Union[str, path_t]:
    """"""
    translations = str.maketrans(
        PATH_ILLEGAL_CHARACTERS, PATH_ILLEGAL_CHARACTERS.__len__() * replacement
    )
    if isinstance(path, str):
        as_path = path_t(path)
    else:
        as_path = path
    parts = as_path.parts
    if parts[0] == as_path.drive + as_path.root:
        output = [parts[0]]
        p_idx = 1
    else:
        output = []
        p_idx = 0
    for part in parts[p_idx:]:
        output.append(part.translate(translations))
    output = path_t(*output)

    if isinstance(path, str):
        return str(output)

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Dict, Optional, Sequence, Tuple

import numpy as nmpy
import scipy.ndimage as spim
import scipy.optimize as spop
from scipy import ndimage as ndimage_t
from scipy.spatial import distance as dstc
from skimage import morphology as mrph
from skimage import segmentation as sgmt

import cell_tracking_BC.task.registration as rgst
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.task.jaccard import PseudoJaccard
from cell_tracking_BC.type.compartment.base import CompartmentReference


array_t = nmpy.ndarray


MAX_N_NUCLEI = 3

_ROUNDED_SQUARE_5 = nmpy.ones((5, 5), dtype=nmpy.bool_)
_ROUNDED_SQUARE_5[0, 0] = False
_ROUNDED_SQUARE_5[0, -1] = False
_ROUNDED_SQUARE_5[-1, 0] = False
_ROUNDED_SQUARE_5[-1, -1] = False


def CorrectBasedOnTemporalCoherence(
    current: array_t,
    previous: array_t,
    /,
    *,
    min_jaccard: float = 0.75,
    max_area_discrepancy: float = 0.25,
    min_cell_area: int = 0,
    time_point: int = None,
) -> Tuple[int, array_t]:
    """
    min_jaccard: Actually, Pseudo-Jaccard
    """
    assert nmpy.issubdtype(current.dtype, nmpy.bool_)

    if time_point is None:
        at_time_point = ""
    else:
        at_time_point = f" at time point {time_point}"

    labeled_current, n_cells_current = mrph.label(
        current, return_num=True, connectivity=1
    )
    labeled_previous, n_cells_previous = mrph.label(
        previous, return_num=True, connectivity=1
    )
    jaccards = _PairwiseJaccards(
        n_cells_previous, n_cells_current, labeled_previous, labeled_current
    )
    c_to_p_links = _CurrentToPreviousLinks(jaccards, min_jaccard)

    frontiers = nmpy.zeros_like(current)
    n_corrections = 0
    for label_current in range(1, n_cells_current + 1):
        labels_previous = c_to_p_links.get(label_current - 1)
        if (labels_previous is not None) and (labels_previous.__len__() > 1):
            labels_previous = nmpy.array(labels_previous) + 1
            context = f"between cells {sorted(labels_previous)} and fused cell {label_current}{at_time_point}"

            where_fused = labeled_current == label_current

            splitted = _SplitLabels(
                where_fused,
                labeled_previous,
                labels_previous,
                max_area_discrepancy,
                min_cell_area,
                context,
            )
            if splitted is None:
                continue

            frontier = nmpy.logical_and(where_fused, splitted == 0)
            current[frontier] = False
            frontiers[frontier] = True
            n_corrections += 1

    return n_corrections, frontiers


def _PairwiseJaccards(
    n_cells_previous: int,
    n_cells_current: int,
    labeled_previous: array_t,
    labeled_current: array_t,
    /,
) -> array_t:
    """"""
    labels_previous = nmpy.fromiter(range(1, n_cells_previous + 1), dtype=nmpy.uint64)
    labels_current = nmpy.fromiter(range(1, n_cells_current + 1), dtype=nmpy.uint64)
    # Note the reversed parameter order in PseudoJaccard since a fusion is a division in reversed time
    _PseudoJaccard = lambda lbl_1, lbl_2: PseudoJaccard(
        labeled_current, labeled_previous, lbl_2, lbl_1
    )

    return dstc.cdist(
        labels_previous[:, None], labels_current[:, None], metric=_PseudoJaccard
    )


def _CurrentToPreviousLinks(
    pairwise_jaccards: array_t, min_jaccard: float, /
) -> Dict[int, Sequence[int]]:
    """"""
    output = {}

    while True:
        row_idc, col_idc = spop.linear_sum_assignment(1.0 - pairwise_jaccards)
        valid_idc = pairwise_jaccards[row_idc, col_idc] > min_jaccard
        if not nmpy.any(valid_idc):
            break

        valid_row_idc = row_idc[valid_idc]
        for col_idx, row_idx in zip(col_idc[valid_idc], valid_row_idc):
            if col_idx in output:
                output[col_idx].append(row_idx)
            else:
                output[col_idx] = [row_idx]

        pairwise_jaccards[valid_row_idc, :] = 0.0

    return output


def _SplitLabels(
    where_fused: array_t,
    labeled_previous: array_t,
    labels_previous: array_t,
    max_area_discrepancy: float,
    min_cell_area: int,
    context: str,
    /,
) -> Optional[array_t]:
    """"""
    output = nmpy.zeros_like(labeled_previous)
    for l_idx, label_previous in enumerate(labels_previous, start=1):
        output[labeled_previous == label_previous] = l_idx

    fused_area = nmpy.count_nonzero(where_fused)
    split_area = nmpy.count_nonzero(output)
    discrepancy = abs(split_area - fused_area) / fused_area
    if discrepancy > max_area_discrepancy:
        LOGGER.warning(
            f"Segmentation correction discarded due to high t-total-area/(t+1)-fused-area discrepancy "
            f"{context}: "
            f"Actual={discrepancy}. Expected<={max_area_discrepancy}."
        )
        return None

    (fused_local, split_local), corners = rgst.InCommonNonZeroRectangles(
        where_fused, output, for_rotation=True
    )

    angle = rgst.RotationInBinary(fused_local, split_local > 0)
    rotated = rgst.RotatedLabeled(split_local, angle)

    split_local = _DilatedWithLabelPreservation(rotated, fused_local, min_cell_area)
    if split_local is None:
        LOGGER.warning(
            f"Segmentation correction discarded due to invalid split regions "
            f"{context}"
        )
        return None

    corner = corners[0]
    rows_gbl, cols_gbl = nmpy.meshgrid(
        range(corner[0], corner[0] + split_local.shape[0]),
        range(corner[1], corner[1] + split_local.shape[1]),
        indexing="ij",
    )
    rows_lcl, cols_lcl = nmpy.indices(split_local.shape)
    valid = nmpy.logical_and(
        nmpy.logical_and(rows_gbl >= 0, cols_gbl >= 0),
        nmpy.logical_and(rows_gbl < output.shape[0], cols_gbl < output.shape[1]),
    )
    output.fill(0)
    output[rows_gbl[valid], cols_gbl[valid]] = split_local[
        rows_lcl[valid], cols_lcl[valid]
    ]

    return output


def _DilatedWithLabelPreservation(
    labeled: array_t, roi: array_t, min_cell_area: int, /
) -> Optional[array_t]:
    """"""
    output = nmpy.ones_like(labeled)

    n_labels = nmpy.amax(labeled).item()
    non_roi = nmpy.logical_not(roi)

    distance_map = spim.distance_transform_edt(labeled != 1)
    for label in range(2, n_labels + 1):
        current_map = spim.distance_transform_edt(labeled != label)
        closer_b_map = current_map < distance_map
        output[closer_b_map] = label
        distance_map[closer_b_map] = current_map[closer_b_map]

    while True:
        intermediate = nmpy.zeros_like(output)
        for label in range(1, n_labels + 1):
            where_label = output == label
            if not nmpy.any(where_label):
                return None
            # TODO: check how erosion behaves on image borders
            eroded = mrph.binary_erosion(where_label)
            intermediate[eroded] = label
        nmpy.copyto(output, intermediate)
        output[non_roi] = 0

        labeled_output, n_output_labels = mrph.label(
            output > 0, return_num=True, connectivity=1
        )
        areas = tuple(
            nmpy.count_nonzero(labeled_output == _lbl)
            for _lbl in range(1, n_output_labels + 1)
        )
        for label, area in enumerate(areas, start=1):
            if area < min_cell_area:
                output[labeled_output == label] = 0
                n_output_labels -= 1
        if n_output_labels == n_labels:
            break

    return output


def AllCompartmentsFromSome(
    *,
    cells_map: array_t = None,
    cytoplasms_map: array_t = None,
    nuclei_map: array_t = None,
) -> Tuple[array_t, array_t, array_t]:
    """
    Maps need not be booleans since they are converted.

    Valid options:
        - cell               => nucleus = cell reference pixel, cytoplasm = cell - nucleus
        - cell, cytoplasm    => nucleus = cell - cytoplasm
        - cell, nucleus      => cytoplasm = cell - nucleus
        - cytoplasm          => cell = filled cytoplasm, nucleus = cell - cytoplasm
        - cytoplasm, nucleus => cell = filled cytoplasm
    """
    if (cells_map is None) and (cytoplasms_map is None):
        raise ValueError("Cytoplasms and cells maps both None")
    if not ((cells_map is None) or (cytoplasms_map is None) or (nuclei_map is None)):
        raise ValueError("Nuclei, cytoplasms and cells maps all not None")

    if cells_map is not None:
        cells_map = cells_map > 0
        _CorrectCellsMap(cells_map)
        _CheckCellsMap(cells_map)
    if cytoplasms_map is not None:
        cytoplasms_map = cytoplasms_map > 0
        _CorrectCytoplasmsMap(cytoplasms_map)
        _CheckCytoplasmsMap(cytoplasms_map)
    if nuclei_map is not None:
        nuclei_map = nuclei_map > 0
        _CorrectNucleiMap(nuclei_map)
        _CheckNucleiMap(nuclei_map)

    if cells_map is None:  # Then cytoplasms_map is not None
        if (nuclei_map is not None) and nmpy.any(cytoplasms_map[nuclei_map]):
            raise ValueError("Cytoplasms and nuclei maps intersect")

        cells_map = _CellsMapFromCytoplasms(cytoplasms_map)

        if nuclei_map is None:
            nuclei_map = _InnerCompartmentsMapFromCellsAndOtherInners(
                cells_map, cytoplasms_map
            )
        else:
            eroded_cells_map = _SafelyErodedCellsMap(cells_map)
            if not nmpy.all(eroded_cells_map[nuclei_map]):
                LOGGER.warning(
                    "Nuclei not restricted to cells, or tangent: Correcting nuclei and cytoplasms maps"
                )
                _FitNucleiInErodedCells(eroded_cells_map, nuclei_map)
                cytoplasms_map = _InnerCompartmentsMapFromCellsAndOtherInners(
                    cells_map, nuclei_map
                )
            else:
                union = nmpy.logical_or(cytoplasms_map, nuclei_map)
                if not nmpy.array_equal(cells_map, union):
                    raise ValueError(
                        "Cytoplasms inner borders do not coincide with nuclei"
                    )

            _CorrectNumberOfNucleiPerCell(cells_map, cytoplasms_map, nuclei_map)
    elif cytoplasms_map is None:  # ...and cells_map is not None
        if nuclei_map is None:
            LOGGER.warning("Synthetic cytoplasms and nuclei maps")
            nuclei_map = _NucleiMapFromCells(cells_map)
            cytoplasms_map = _InnerCompartmentsMapFromCellsAndOtherInners(
                cells_map, nuclei_map
            )
        else:  # cells_map is not None, cytoplasms_map is None, and nuclei_map is not None
            # /!\ Almost code duplication
            eroded_cells_map = _SafelyErodedCellsMap(cells_map)
            if not nmpy.all(eroded_cells_map[nuclei_map]):
                LOGGER.warning(
                    "Nuclei not restricted to cells, or tangent: Correcting nuclei map"
                )
                _FitNucleiInErodedCells(eroded_cells_map, nuclei_map)

            cytoplasms_map = _InnerCompartmentsMapFromCellsAndOtherInners(
                cells_map, nuclei_map
            )

            _CorrectNumberOfNucleiPerCell(cells_map, cytoplasms_map, nuclei_map)
    else:  # cells_map is not None and cytoplasms_map is not None (hence, nuclei_map is None)
        _CheckNumberOfCytoplasmsPerCell(cells_map, cytoplasms_map)

        from_cytoplasm = _CellsMapFromCytoplasms(cytoplasms_map)
        if not nmpy.array_equal(from_cytoplasm, cells_map):
            raise ValueError("Cytoplasms outer borders do not coincide with cells")

        nuclei_map = _InnerCompartmentsMapFromCellsAndOtherInners(
            cells_map, cytoplasms_map
        )

    return cells_map, cytoplasms_map, nuclei_map


def _CorrectCellsMap(map_: array_t, /) -> None:
    """"""
    mrph.remove_small_holes(map_, area_threshold=1000, out=map_)
    _RemoveTinyCompartments(map_, map_, "cells")


def _CorrectCytoplasmsMap(map_: array_t, /) -> None:
    """"""
    filled = _CellsMapFromCytoplasms(map_)
    _RemoveTinyCompartments(map_, filled, "cytoplasms")


def _CorrectNucleiMap(map_: array_t, /) -> None:
    """"""
    mrph.remove_small_holes(map_, area_threshold=1000, out=map_)


def _RemoveTinyCompartments(
    to_be_corrected: array_t, to_be_labeled: array_t, plural_name: str, /
) -> None:
    """"""
    l_map, n_compartments = ndimage_t.label(to_be_labeled)  # connectivity=1
    modified = False
    for label in range(1, n_compartments + 1):
        current = l_map == label
        distance_map = ndimage_t.distance_transform_edt(current)
        if nmpy.amax(distance_map) < 2.0:
            to_be_corrected[current] = False
            modified = True

    if modified:
        LOGGER.warning(f"Tiny {plural_name} were removed")


def _CellsMapFromCytoplasms(map_: array_t, /) -> array_t:
    """"""
    return ndimage_t.binary_fill_holes(map_)


def _NucleiMapFromCells(map_: array_t, /) -> array_t:
    """"""
    output = nmpy.zeros_like(map_)

    cells_l_map, n_cells = ndimage_t.label(map_)  # connectivity=1
    for label in range(1, n_cells + 1):
        output[tuple(CompartmentReference(cells_l_map == label))] = True

    return output


def _InnerCompartmentsMapFromCellsAndOtherInners(
    cells_map: array_t, inners_map: array_t, /
) -> array_t:
    """"""
    output = cells_map.copy()
    output[inners_map] = False

    return output


def _FitNucleiInErodedCells(eroded_cells_map: array_t, nuclei_map: array_t, /) -> None:
    """"""
    nuclei_map[nmpy.logical_not(eroded_cells_map)] = False

    cells_l_map, n_cells = ndimage_t.label(eroded_cells_map)  # connectivity=1
    for label in range(1, n_cells + 1):
        cell_l_map = cells_l_map == label
        if not nmpy.any(nuclei_map[cell_l_map]):
            nuclei_map[tuple(CompartmentReference(cell_l_map))] = True


def _CorrectNumberOfNucleiPerCell(
    cells_map: array_t, cytoplasms_map: array_t, nuclei_map: array_t, /
) -> None:
    """"""
    cells_l_map, n_cells = ndimage_t.label(cells_map)  # connectivity=1
    nuclei_l_map, _ = ndimage_t.label(nuclei_map)  # connectivity=1

    for label in range(1, n_cells + 1):
        cell_l_map = cells_l_map == label
        nuclei_inside_cell = nuclei_l_map[cell_l_map]
        nuclei_labels = nmpy.unique(nuclei_inside_cell)
        # Do not use nuclei_labels[1:] since nuclei_labels[0] is not necessarily zero
        nuclei_labels = nuclei_labels[nuclei_labels > 0]
        n_nuclei = nuclei_labels.size
        if n_nuclei == 0:
            LOGGER.warning(
                f"{n_nuclei}: Invalid number of nuclei in cytoplasm. Expected=1 or 2. "
                f"Correcting nuclei and cytoplasms maps"
            )
            reference = tuple(CompartmentReference(cell_l_map))
            nuclei_map[reference] = True
            cytoplasms_map[reference] = False
        elif n_nuclei > MAX_N_NUCLEI:
            LOGGER.warning(
                f"{n_nuclei}: Invalid number of nuclei in cytoplasm. Expected>=1, <={MAX_N_NUCLEI}. "
                f"Correcting nuclei and cytoplasms maps"
            )
            nuclei_labels = sorted(
                nuclei_labels,
                key=lambda _lbl: nmpy.count_nonzero(nuclei_inside_cell == _lbl),
                reverse=True,
            )
            for too_small in nuclei_labels[MAX_N_NUCLEI:]:
                extra_nucleus = nuclei_l_map == too_small
                nuclei_map[extra_nucleus] = False
                cytoplasms_map[extra_nucleus] = True


def _CheckCellsMap(map_: array_t, /) -> None:
    """"""
    _CheckCompartmentsHoles(map_, "cell", False, 0, 0)


def _CheckCytoplasmsMap(map_: array_t, /) -> None:
    """"""
    _CheckCompartmentsHoles(map_, "cytoplasm", True, 1, MAX_N_NUCLEI)
    # cleared = sgmt.clear_border(cytoplasm)
    # labeled_components, n_components = mrph.label(
    #     cleared, return_num=True, connectivity=1
    # )
    # for label in range(1, n_components + 1):
    #     # Actually, n_holes = n_true_holes + background
    #     _, n_holes = mrph.label(
    #         labeled_components != label, return_num=True, connectivity=2
    #     )
    #     if n_holes < 2:
    #         # print(label)
    #         # import matplotlib.pyplot as p
    #         # p.matshow(a)
    #         # p.matshow(labeled_components)
    #         # p.matshow(b)
    #         # p.show()
    #         raise ValueError("Cytoplasm w/o holes")


def _CheckNumberOfCytoplasmsPerCell(
    cells_map: array_t, cytoplasms_map: array_t, /
) -> None:
    """"""
    cells_l_map, n_cells = ndimage_t.label(cells_map)  # connectivity=1
    cytoplasms_l_map, _ = ndimage_t.label(cytoplasms_map)  # connectivity=1

    for label in range(1, n_cells + 1):
        cell_l_map = cells_l_map == label
        cytoplasms_labels = nmpy.unique(cytoplasms_l_map[cell_l_map])
        # Do not use cytoplasms_labels[1:] since cytoplasms_labels[0] is not necessarily zero
        cytoplasms_labels = cytoplasms_labels[cytoplasms_labels > 0]
        n_cytoplasms = cytoplasms_labels.size
        if n_cytoplasms != 1:
            raise ValueError(
                f"{n_cytoplasms}: Invalid number of cytoplasms in cell. Expected=1."
            )


def _CheckNucleiMap(map_: array_t, /) -> None:
    """"""
    _CheckCompartmentsHoles(map_, "nucleus", False, 0, 0)


def _CheckCompartmentsHoles(
    compartments_map: array_t,
    name: str,
    expect_on_border: bool,
    min_n_holes: int,
    max_n_holes: int,
    /,
) -> None:
    """"""
    if expect_on_border:
        cleared = sgmt.clear_border(compartments_map)
    else:
        cleared = compartments_map

    # +1 for background
    min_n_holes += 1
    max_n_holes += 1

    compartments_l_map, n_compartments = mrph.label(
        cleared, return_num=True, connectivity=1
    )
    for label in range(1, n_compartments + 1):
        # Actually, n_holes = n_true_holes + 1 for background
        _, n_holes = mrph.label(
            compartments_l_map != label, return_num=True, connectivity=2
        )
        if (n_holes < min_n_holes) or (n_holes > max_n_holes):
            # print(label)
            # import matplotlib.pyplot as p
            # p.matshow(a)
            # p.matshow(labeled_components)
            # p.matshow(b)
            # p.show()
            raise ValueError(
                f"{n_holes}: Invalid number of holes in {name}. Expected>={min_n_holes-1}, <={max_n_holes-1}"
            )


def _SafelyErodedCellsMap(map_: array_t, /) -> array_t:
    """"""
    output = ndimage_t.binary_erosion(map_, structure=_ROUNDED_SQUARE_5)
    if nmpy.any(output):
        return output

    raise ValueError("Cell too small to contain nuclei")



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from enum import Enum as enum_t
from typing import ClassVar, Dict, Sequence, Tuple, Union

import numpy as nmpy
import skimage.morphology as mrph

from cell_tracking_BC.type.compartment.base import compartment_t, compartment_id_t
from cell_tracking_BC.type.compartment.cytoplasm import cytoplasm_t
from cell_tracking_BC.type.compartment.nucleus import nucleus_t


array_t = nmpy.ndarray


class state_e(enum_t):
    unknown = None
    discarded = -1
    living = 1
    dividing = 2
    dead = 0

    def IsActive(self, /, *, strict_mode: bool = False) -> bool:
        """"""
        if strict_mode:
            return self not in (state_e.dead, state_e.discarded, state_e.unknown)

        return self not in (state_e.dead, state_e.discarded)

    def IsAlive(self, /, *, strict_mode: bool = False) -> bool:
        """"""
        if strict_mode:
            return self in (state_e.living, state_e.dividing)

        return self in (state_e.living, state_e.dividing, state_e.unknown)


@dtcl.dataclass(repr=False, eq=False)
class cell_t(compartment_t):

    NEXT_AVAILABLE_LABEL: ClassVar[int] = 1

    label: int = None
    state: state_e = state_e.unknown
    cytoplasm: cytoplasm_t = None
    nuclei: Tuple[nucleus_t, ...] = None
    # And inherited from compartment_t (among others): features

    def __post_init__(self) -> None:
        """"""
        self.__class__.NEXT_AVAILABLE_LABEL += 1

    @classmethod
    def NewFromMap(cls, _: array_t, /, *, is_plain: bool = True) -> compartment_t:
        """"""
        raise RuntimeError(
            f"{cell_t.NewFromMap.__name__}: Not meant to be called from class {cell_t.__name__}; "
            f"Use {cell_t.NewFromMaps.__name__} instead"
        )

    @classmethod
    def NewFromMaps(
        cls,
        /,
        *,
        cell_map: array_t = None,
        nucleus_map: array_t = None,
    ) -> cell_t:
        """
        cell_map: Defined optional for convenience (calling with kwargs), but never None in fact
        Valid options: see CellAndCytoplasmMapsFromCombinations
        """
        cell = compartment_t.NewFromMap(cell_map)
        if nucleus_map is None:
            cytoplasm = None
            nuclei = None
        else:
            cytoplasm = cytoplasm_t.NewFromMap(
                nmpy.logical_xor(cell_map, nucleus_map), is_plain=False
            )
            labeled_nuclei, n_nuclei = mrph.label(
                nucleus_map, return_num=True, connectivity=1
            )
            nuclei = tuple(
                nucleus_t.NewFromMap(labeled_nuclei == _lbl)
                for _lbl in range(1, n_nuclei + 1)
            )

        instance = cls(
            label=cls.NEXT_AVAILABLE_LABEL,
            cytoplasm=cytoplasm,
            nuclei=nuclei,
            centroid=cell.centroid,
            bb_slices=cell.bb_slices,
            touches_border=cell.touches_border,
            map_stream=cell.map_stream,
            area=cell.area,
        )

        return instance

    @property
    def compartments(
        self,
    ) -> Dict[
        Union[compartment_id_t, str], Union[compartment_t, Sequence[compartment_t]]
    ]:
        """"""
        if self.cytoplasm is None:
            return {compartment_id_t.CELL: self}

        return {
            compartment_id_t.CELL: self,
            compartment_id_t.CYTOPLASM: self.cytoplasm,
            compartment_id_t.NUCLEUS: self.nuclei,
        }

    def Map(
        self,
        shape: Sequence[int],
        /,
        *,
        as_boolean: bool = False,
        with_labels: bool = False,
        margin: float = None,
    ) -> array_t:
        """
        with_labels: cytoplasm will be marked with label 1, nuclei with subsequent labels
        /!\ If with_labels, no margin should be requested. Indeed, what should be done with the nucleus map then?
        Apply the same margin or not? Or the method should take an additional margin parameter for the nucleus.
        """
        output = super().Map(shape, as_boolean=as_boolean, margin=margin)

        if with_labels:
            if as_boolean:
                raise ValueError('"with_labels" and "as_boolean" cannot be both True')

            if self.nuclei is not None:
                for n_idx, nucleus in enumerate(self.nuclei, start=2):
                    map_ = nucleus.Map(shape, as_boolean=True)
                    output[map_] = n_idx

        return output

    def NucleiMap(
        self,
        shape: Sequence[int],
        /,
        *,
        as_boolean: bool = False,
        margin: float = None,
    ) -> array_t:
        """"""
        output, one = compartment_t.EmptyMap(shape, as_boolean=as_boolean)

        for nucleus in self.nuclei:
            output[nucleus.Map(shape, as_boolean=True, margin=margin)] = one

        return output

    def __hash__(self) -> int:
        """
        Note that (from Python documentation:
            if it defines __eq__() but not __hash__(), its instances will not be usable as items in hashable collections
        """
        return hash((self.label, tuple(self.centroid), self.map_stream))

    def __str__(self) -> str:
        """"""
        if self.nuclei is None:
            n_nuclei = 0
        else:
            n_nuclei = self.nuclei.__len__()

        return super().__str__() + f"    Label: {self.label}\n    Nuclei: {n_nuclei}\n"



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional, Sequence, Tuple, Union

import numpy as nmpy

import cell_tracking_BC.in_out.graphics.generic.d_3 as thrd
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t, axes_3d_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.type.analysis import analysis_t
from cell_tracking_BC.type.compartment.cell import cell_t
from cell_tracking_BC.type.track.single.unstructured import unstructured_track_t


_LOW_AFFINITY = 0.75
_MATPLOTLIB_COLORS = ("b", "g", "r", "c", "m", "y", "k")


def ShowTracking2D(
    analysis: analysis_t,
    /,
    *,
    with_segmentation: bool = True,
    with_cell_labels: bool = True,
    with_track_labels: bool = True,
    with_ticks: bool = True,
    with_colorbar: bool = True,
    mode: str = "forking",
    figure_name: str = "tracking-2D",
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[figure_t]:
    """"""
    output = None

    explorer = dbe.t_explorer_2d_t.NewForSequence(
        analysis,
        dbe,
        with_segmentation=with_segmentation,
        with_cell_labels=with_cell_labels,
        with_track_labels=with_track_labels,
        with_ticks=with_ticks,
        with_colorbar=with_colorbar,
        mode=mode,
    )

    explorer.figure.Archive(name=figure_name)
    if prepare_only:
        output = explorer
    else:
        explorer.figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def ShowTracking3D(
    analysis: analysis_t,
    /,
    *,
    with_track_labels: bool = True,
    with_cell_labels: bool = True,
    figure_name: str = "tracking-3D",
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[figure_t]:
    """"""
    output = None

    figure, axes = dbe.figure_3d_t.NewFigureAndAxes(title=figure_name)
    axes: axes_3d_t
    colormap = axes.AddColormapFromMilestones(
        "Tracking Affinity", ((0.0, "black"), (0.75, "red"), (1.0, "blue"))
    )

    thrd.PlotFirstFrameAsFloor(
        analysis.sequence[analysis.sequence.plot_channel][0], axes
    )

    time_scaling = axes.__class__.TimeScaling(
        analysis.sequence.shape, analysis.sequence.length
    )
    for t_idx, track in enumerate(analysis.tracks):
        low_affinities = tuple(
            _ffn < _LOW_AFFINITY for _ffn in track.tracking_affinities
        )
        low_fraction = nmpy.count_nonzero(low_affinities) / (
            0.3 * low_affinities.__len__()
        )
        color = colormap(1.0 - min(1.0, low_fraction))

        for cells, start_time_point, label in track.PiecesIterator():
            n_cells = cells.__len__()
            # Use *cell_labels instead of cell_labels in case with_cell_labels is False, which then makes the labels
            # accessible through cell_labels[0].
            rows, cols, times, *cell_labels = _AsRowsColsTimes(
                cells, start_time_point, with_labels=with_cell_labels
            )
            times = tuple(time_scaling * _tme for _tme in times)

            axes.PlotLines(rows, cols, times, color=color)

            if with_cell_labels:
                after_last_idx = n_cells - 1 if label is None else n_cells
                for c_idx in range(after_last_idx):
                    axes.PlotText(
                        rows[c_idx],
                        cols[c_idx],
                        times[c_idx],
                        str(cell_labels[0][c_idx]),
                        fontsize="x-small",
                        color=color,
                    )
            if with_track_labels and (label is not None):
                axes.PlotText(
                    rows[-1],
                    cols[-1],
                    times[-1] + 0.25,
                    str(label),
                    fontsize="x-small",
                    color=color,
                )

    axes.SetTimeAxisProperties(analysis.sequence.length - 1)

    figure.Archive(name=figure_name)
    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def ShowInvalidTracking2D(
    analysis: analysis_t,
    /,
    *,
    with_cell_labels: bool = True,
    figure_name: str = "invalid-tracking-2D",
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[figure_t]:
    """"""
    output = None

    invalids = [_rcd[0] for _rcd in analysis.tracks.invalids]
    if invalids.__len__() == 0:
        return

    figure, axes = dbe.figure_2d_t.NewFigureAndAxes(title=figure_name)
    axes: axes_2d_t

    height = analysis.sequence.shape[0]
    current_base_pos = 0.0
    colors = _MATPLOTLIB_COLORS
    for t_idx, track in enumerate(invalids):
        color_idx = t_idx % colors.__len__()

        rows, cols = zip(*(_cll.centroid.tolist() for _cll in track))
        positions = tuple(_row + _col * height for _row, _col in zip(rows, cols))
        min_pos = min(positions)
        max_pos = max(positions)
        if max_pos == min_pos:
            pos_scaling = 1.0
        else:
            pos_scaling = 1.0 / (max_pos - min_pos)

        if isinstance(track, unstructured_track_t):
            for time_point, cells, has_leaf in track.multi_segments_iterator:
                n_cells = cells.__len__()
                rows, cols = zip(*(_cll.centroid.tolist() for _cll in cells))
                positions = nmpy.fromiter(
                    (_row + _col * height for _row, _col in zip(rows, cols)),
                    nmpy.float64,
                )
                positions = pos_scaling * (positions - min_pos) + current_base_pos
                times = tuple(time_point + _idx for _idx in range(n_cells))

                axes.PlotLines(times, positions, colors[color_idx])

                if with_cell_labels:
                    after_last_idx = n_cells if has_leaf else n_cells - 1
                    labels = tuple(_cll.label for _cll in cells)
                    for c_idx in range(after_last_idx):
                        axes.PlotText(
                            times[c_idx],
                            positions[c_idx],
                            str(labels[c_idx]),
                            fontsize="x-small",
                            color=colors[color_idx],
                        )
        else:
            for cells, start_time_point, label in track.PiecesIterator():
                n_cells = cells.__len__()
                # Use *cell_labels instead of cell_labels in case with_cell_labels is False, which then makes the labels
                # accessible through cell_labels[0].
                rows, cols, times, *cell_labels = _AsRowsColsTimes(
                    cells, start_time_point, with_labels=with_cell_labels
                )
                positions = nmpy.fromiter(
                    (_row + _col * height for _row, _col in zip(rows, cols)),
                    nmpy.float64,
                )
                positions = pos_scaling * (positions - min_pos) + current_base_pos

                axes.PlotLines(times, positions, colors[color_idx])

                if with_cell_labels:
                    after_last_idx = n_cells - 1 if label is None else n_cells
                    for c_idx in range(after_last_idx):
                        axes.PlotText(
                            times[c_idx],
                            positions[c_idx],
                            str(cell_labels[0][c_idx]),
                            fontsize="x-small",
                            color=colors[color_idx],
                        )
                if label is not None:
                    axes.PlotText(
                        times[-1] + 0.25,
                        positions[-1],
                        str(label),
                        fontsize = "x-small",
                        color = colors[color_idx],
                    )
        current_base_pos += 1.25

    axes.SetTimeAxisProperties(analysis.sequence.length - 1)

    figure.Archive(name=figure_name)
    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def ShowInvalidTracking3D(
    analysis: analysis_t,
    /,
    *,
    with_cell_labels: bool = True,
    figure_name: str = "invalid-tracking-3D",
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[figure_t]:
    """"""
    output = None

    invalids = [_rcd[0] for _rcd in analysis.tracks.invalids]
    if invalids.__len__() == 0:
        return

    figure, axes = dbe.figure_3d_t.NewFigureAndAxes(title=figure_name)
    axes: axes_3d_t

    thrd.PlotFirstFrameAsFloor(
        analysis.sequence[analysis.sequence.plot_channel][0], axes
    )

    colors = _MATPLOTLIB_COLORS
    for t_idx, track in enumerate(invalids):
        color_idx = t_idx % colors.__len__()

        if isinstance(track, unstructured_track_t):
            for time_point, cells, has_leaf in track.multi_segments_iterator:
                n_cells = cells.__len__()
                rows, cols = tuple(zip(*(_cll.centroid.tolist() for _cll in cells)))
                times = tuple(time_point + _idx for _idx in range(n_cells))

                axes.PlotLines(rows, cols, times, colors[color_idx])

                if with_cell_labels:
                    after_last_idx = n_cells if has_leaf else n_cells - 1
                    labels = tuple(_cll.label for _cll in cells)
                    for c_idx in range(after_last_idx):
                        axes.PlotText(
                            rows[c_idx],
                            cols[c_idx],
                            times[c_idx],
                            str(labels[c_idx]),
                            fontsize="x-small",
                            color=colors[color_idx],
                        )
        else:
            for cells, start_time_point, label in track.PiecesIterator():
                n_cells = cells.__len__()
                # Use *cell_labels instead of cell_labels in case with_cell_labels is False, which then makes the labels
                # accessible through cell_labels[0].
                rows, cols, times, *cell_labels = _AsRowsColsTimes(
                    cells, start_time_point, with_labels=with_cell_labels
                )

                axes.PlotLines(rows, cols, times, colors[color_idx])

                if with_cell_labels:
                    after_last_idx = n_cells - 1 if label is None else n_cells
                    for c_idx in range(after_last_idx):
                        axes.PlotText(
                            rows[c_idx],
                            cols[c_idx],
                            times[c_idx],
                            str(cell_labels[0][c_idx]),
                            fontsize="x-small",
                            color=colors[color_idx],
                        )
                if label is not None:
                    axes.PlotText(
                        rows[-1],
                        cols[-1],
                        times[-1] + 0.25,
                        str(label),
                        fontsize = "x-small",
                        color = colors[color_idx],
                    )

    axes.SetTimeAxisProperties(analysis.sequence.length - 1)

    figure.Archive(name=figure_name)
    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def _AsRowsColsTimes(
    cells: Sequence[cell_t], start_time_point: int, /, *, with_labels: bool = False
) -> Union[
    Tuple[Tuple[float, ...], Tuple[float, ...], Tuple[int, ...]],
    Tuple[Tuple[float, ...], Tuple[float, ...], Tuple[int, ...], Tuple[int, ...]],
]:
    """"""
    n_cells = cells.__len__()

    rows, cols = tuple(zip(*(_cll.centroid.tolist() for _cll in cells)))
    times = tuple(range(start_time_point, start_time_point + n_cells))

    if with_labels:
        labels = tuple(_cll.label for _cll in cells)
        return rows, cols, times, labels

    return rows, cols, times



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import tempfile as temp
import time
from pathlib import Path as path_t
from typing import Any, Dict, Sequence, Tuple, Union

import matplotlib.pyplot as pypl
import numpy as nmpy
import skimage.measure as msre
import skimage.transform as trfm
import tifffile as tiff

from cell_tracking_BC.in_out.graphics.dbe.matplotlib.style import CellAnnotationStyle
from cell_tracking_BC.type.analysis import analysis_t
from cell_tracking_BC.type.compartment.base import AdjustedMapWithMargin
from cell_tracking_BC.type.compartment.cell import cell_t, state_e


array_t = nmpy.ndarray


def SaveSequenceAnnotations(
    path: Union[str, path_t],
    analysis: analysis_t,
    /,
    *,
    cell_margin: float = -50.0,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if path.is_dir():
        path /= "sequence_annotations.tiff"

    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    track_labels = {_tpt: {} for _tpt in range(analysis.sequence.length)}
    for track in analysis.tracks:
        root_time_point = track.topologic_root_time_point
        for cells, label in track.LabeledThreadIterator(topologic_mode=True):
            for time_point, cell in enumerate(cells, start=root_time_point):
                if cell.state is state_e.discarded:
                    # Leave "-" at the end to allow numerical sorting
                    text = f"{label}-"
                elif cell.state is state_e.dead:
                    # Leave "x" at the end to allow numerical sorting
                    text = f"{label}x"
                elif cell.state is state_e.dividing:
                    # Leave "x" at the end to allow numerical sorting
                    text = f"{label}y"
                else:
                    text = str(label)
                if cell in track_labels[time_point]:
                    track_labels[time_point][cell][0].append(text)
                else:
                    track_labels[time_point][cell] = ([text], cell.centroid)

    backend = pypl.get_backend()
    pypl.switch_backend("Agg")
    figure, axes = pypl.subplots(facecolor="w")
    canvas = figure.canvas
    renderer = canvas.get_renderer()

    annotated = None
    frame_shape = analysis.sequence.shape
    for f_idx in range(analysis.sequence.length):
        if f_idx % 20 == 0:
            print(f"    Writing frame {f_idx} @ {time.ctime()}...")

        axes.clear()
        axes.set_xlim(left=0, right=frame_shape[1] - 1)
        axes.set_ylim(bottom=0, top=frame_shape[0] - 1)
        axes.xaxis.set_visible(False)
        axes.yaxis.set_visible(False)
        axes.set_facecolor("k")
        axes.set_position((0.0, 0.0, 1.0, 1.0))

        description = _CellDescriptions(
            analysis.segmentations[f_idx].cells,
            analysis.sequence.shape,
            margins=(0, cell_margin),
        )
        for contours in description["cell_contours"].values():
            for contour in contours:
                axes.plot(
                    contour[:, 1],
                    contour[:, 0],
                    linestyle=":",
                    color=(0.0, 0.8, 0.8, 0.3),
                )

        for labels, position in track_labels[f_idx].values():
            if (n_labels := labels.__len__()) > 1:
                text = f"{min(labels)}\n(+{n_labels-1})"
            else:
                text = str(labels[0])
            style = CellAnnotationStyle(False, n_labels > 1)
            axes.annotate(
                text,
                nmpy.flipud(position),
                ha="center",
                va="center",
                **style,
            )
        canvas.draw()

        content = nmpy.array(renderer.buffer_rgba())[1:, 1:, :3]
        if annotated is None:
            annotated = nmpy.empty(
                (*content.shape, analysis.sequence.length), dtype=nmpy.uint8
            )
        annotated[..., f_idx] = content

    pypl.close(fig=figure)  # To prevent remaining caught in event loop
    pypl.switch_backend(backend)

    # row_slice, col_slice = BoundingBoxSlices(annotated)
    # annotated = annotated[row_slice, col_slice, :, :]
    annotated = trfm.resize(
        annotated,
        (*analysis.sequence.shape, 3, analysis.sequence.length),
        preserve_range=True,
    )
    annotated = annotated.astype(nmpy.uint8, copy=False)
    annotated = nmpy.moveaxis(annotated, (0, 1, 2, 3), (2, 3, 1, 0))
    annotated = annotated[:, nmpy.newaxis, :, :, :]

    tiff.imwrite(
        str(path),
        annotated,
        photometric="rgb",
        compression="deflate",
        planarconfig="separate",
        metadata={"axes": "XYZCT"},
    )


def _CellDescriptions(
    cells: Sequence[cell_t],
    shape: Tuple[int, int],
    /,
    *,
    margins: Sequence[float] = None,
) -> Dict[str, Any]:
    """"""
    output = {"cell_labels": tuple((_cll.label, _cll.centroid) for _cll in cells)}

    contours = {}
    if margins is None:
        margins = (0.0,)
    for margin in margins:
        for_margin = []
        for cell in cells:
            if (margin is None) or (margin <= 0.0):
                bb_map = nmpy.pad(cell.BBMap().astype(nmpy.int8), 1)
                if (margin is not None) and (margin < 0.0):
                    bb_map = AdjustedMapWithMargin(bb_map, cell.area, margin)
                around_cell = _LargestContourOfMap(bb_map)
                around_cell += (
                    cell.bb_slices[0].start - 1,
                    cell.bb_slices[1].start - 1,
                )
            else:
                cell_map = cell.Map(shape, margin=margin)
                around_cell = _LargestContourOfMap(cell_map)
            for_margin.append(around_cell)
        contours[margin] = tuple(for_margin)
    output["cell_contours"] = contours

    return output


def _LargestContourOfMap(map_: array_t, /) -> array_t:
    """"""
    contours = msre.find_contours(map_, level=0.5)
    if (n_contours := contours.__len__()) > 1:
        lengths = tuple(_ctr.shape[0] for _ctr in contours)
        largest = lengths.index(max(lengths))

        return contours[largest]
    elif n_contours > 0:
        return contours[0]

    raise ValueError("Cell map has no 0.5-isocontour")



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from typing import Callable, List, Optional, Sequence, Tuple, Union

import numpy as nmpy
import scipy.ndimage as ndimage_t
import tensorflow.keras.models as modl

import cell_tracking_BC.in_out.text.progress as prgs
from cell_tracking_BC.task.segmentation.frame import (
    AllCompartmentsFromSome as AllFrameCompartmentsFromSome,
)
from cell_tracking_BC.task.segmentation.frame import (
    CorrectBasedOnTemporalCoherence as CorrectFrameBasedOnTemporalCoherence,
)


array_t = nmpy.ndarray
processing_h = Callable[[array_t], array_t]


def SegmentationsWithTFNetwork(
    frames: Sequence[array_t],
    network_path: Union[str, path_t],
    /,
    *,
    threshold: float = 0.9,
    PreProcessed: processing_h = None,
    PostProcessed: processing_h = None,
) -> Tuple[Sequence[array_t], Sequence[array_t]]:
    """
    PostProcessed: Could be used to clear border objects. However, since one might want to segment cytoplasms and
    nuclei, clearing border objects here could lead to clearing a cytoplasm while keeping its nucleus. Consequently,
    clearing border objects here, i.e. independently for each segmentation task, is not appropriate.
    """
    output_sgm = []
    output_prd = []

    if PreProcessed is not None:
        frames = tuple(PreProcessed(_frm) for _frm in frames)
    if PostProcessed is None:
        PostProcessed = lambda _prm: _prm

    frames = nmpy.array(frames, dtype=nmpy.float32)
    if frames.ndim == 3:
        frames = nmpy.expand_dims(frames, axis=3)

    network = modl.load_model(network_path)
    # network.summary()
    predictions = network.predict(frames, verbose=1)
    # First dimension is time (needs to be removed for single frame reshape below), last dimension is channels
    # (equal to one, thus removed).
    shape = network.layers[0].input_shape[0][1:-1]

    for t_idx, prediction in enumerate(predictions):
        reshaped = nmpy.reshape(prediction, shape)
        segmentation = reshaped > threshold
        post_processed = PostProcessed(segmentation)
        if nmpy.amax(post_processed.astype(nmpy.uint8)) == 0:
            raise ValueError(f"{t_idx}: Empty segmentation")

        output_prd.append(reshaped)
        output_sgm.append(post_processed)

    return output_sgm, output_prd


def InputSizeOfTFNetwork(
    network_path: Union[str, path_t],
    /,
) -> Sequence[int]:
    """
    Input size for single frames = network input layer shape with first (time, which appears as None in network summary)
    and last dimensions (channels, but only one here) removed.
    """
    network = modl.load_model(network_path)
    layer = network.get_layer(index=0)

    return layer.input.shape[1:-1]


def FillTemporalGaps(frames: Sequence[array_t], extent: Optional[int], /) -> None:
    """"""
    if (extent is None) or (extent <= 0):
        return

    volume = nmpy.dstack(frames)
    line = nmpy.ones((1, 1, extent + 1), dtype=nmpy.bool_)
    # TODO: choose between skimage.morphology and ndimage_t versions (same remark everywhere morph. math. is used)
    ndimage_t.binary_closing(volume, structure=line, origin=(0, 0, 0), output=volume)

    # Exclude first and last frames to avoid border effects
    for f_idx in range(1, frames.__len__() - 1):
        frames[f_idx][...] = volume[..., f_idx]


def CorrectBasedOnTemporalCoherence(
    frames: Sequence[array_t],
    /,
    *,
    min_jaccard: float = 0.75,
    max_area_discrepancy: float = 0.25,
    min_cell_area: int = 0,
) -> Sequence[array_t]:
    """
    Actually, Pseudo-Jaccard
    """
    assert nmpy.issubdtype(frames[0].dtype, nmpy.bool_)

    output = frames.__len__() * [nmpy.zeros_like(frames[0])]

    base_description = "Segmentation Correction(s) "
    n_corrections = 0
    with prgs.ProgressDesign() as progress:
        progress_context = prgs.progress_context_t(
            progress,
            frames.__len__() - 1,
            first=1,
            description=base_description + "0",
        )
        for f_idx in progress_context.elements:
            n_new_corrections, output[f_idx] = CorrectFrameBasedOnTemporalCoherence(
                frames[f_idx],
                frames[f_idx - 1],
                min_jaccard=min_jaccard,
                max_area_discrepancy=max_area_discrepancy,
                min_cell_area=min_cell_area,
                time_point=f_idx,
            )
            n_corrections += n_new_corrections
            progress_context.UpdateDescription(base_description + str(n_corrections))

    return output


def AllCompartmentsFromSome(
    *,
    cells_maps: Sequence[array_t] = None,
    cytoplasms_maps: Sequence[array_t] = None,
    nuclei_maps: Sequence[array_t] = None,
) -> Tuple[List[array_t], List[array_t], List[array_t]]:
    """
    Valid options: see AllFrameCompartmentsFromSome
    """
    output_cells = []
    output_cytoplasms = []
    output_nuclei = []

    lengths = tuple(
        0 if _mps is None else _mps.__len__()
        for _mps in (cells_maps, cytoplasms_maps, nuclei_maps)
    )
    if ((max_length := max(lengths)) == 0) or (
        min(_lgt for _lgt in lengths if _lgt > 0) != max_length
    ):
        raise ValueError(
            f"{lengths}: Compartments maps with different lengths, or all empty"
        )
    length = lengths[0]

    if cells_maps is None:
        cells_maps = length * [None]
    if cytoplasms_maps is None:
        cytoplasms_maps = length * [None]
    if nuclei_maps is None:
        nuclei_maps = length * [None]
    for cells_map, cytoplasms_map, nuclei_map in zip(
        cells_maps, cytoplasms_maps, nuclei_maps
    ):
        compartments = AllFrameCompartmentsFromSome(
            cells_map=cells_map,
            cytoplasms_map=cytoplasms_map,
            nuclei_map=nuclei_map,
        )
        output_cells.append(compartments[0])
        output_cytoplasms.append(compartments[1])
        output_nuclei.append(compartments[2])

    return output_cells, output_cytoplasms, output_nuclei



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as nmpy
import skimage.morphology as mrph
from scipy.sparse import csr_array as s_array_t

from cell_tracking_BC.type.compartment.base import compartment_id_t
from cell_tracking_BC.type.compartment.cell import cell_t


array_t = nmpy.ndarray
frame_h = s_array_t
inner_outer_associations_h = Dict[int, Union[int, Tuple[int, int]]]
CellIssues_h = Callable[[int, array_t, dict], Optional[Union[str, Sequence[str]]]]


_AsArray = s_array_t.toarray


@dtcl.dataclass(repr=False, eq=False)
class segmentation_t:

    shape: Tuple[int, int] = None
    _cells_map: frame_h = None  # dtype=bool; Never None after instantiation
    _nuclei_map: frame_h = None  # dtype=bool
    cells: List[cell_t] = dtcl.field(init=False, default_factory=list)
    invalid_cells: List[Tuple[int, Union[str, Sequence[str]]]] = dtcl.field(
        init=False, default_factory=list
    )

    @classmethod
    def NewFromCellsMap(
        cls,
        cells_map: array_t,
        /,
        *,
        nuclei_map: array_t = None,
    ) -> segmentation_t:
        """"""
        shape = cells_map.shape
        cells_map = s_array_t(cells_map)
        if nuclei_map is not None:
            nuclei_map = s_array_t(nuclei_map)

        return cls(shape=shape, _cells_map=cells_map, _nuclei_map=nuclei_map)

    @property
    def has_nuclei(self) -> bool:
        """"""
        return self._nuclei_map is not None

    @property
    def cells_map(self) -> array_t:
        """"""
        return _AsArray(self._cells_map)

    @property
    def cytoplasms_map(self) -> array_t:
        """"""
        output = (
            self.cells_map.copy()
        )  # Copy: only in case "cells_map" becomes an array one day
        output[self.nuclei_map] = False

        return output

    @property
    def nuclei_map(self) -> array_t:
        """"""
        return _AsArray(self._nuclei_map)

    # TODO: regularly check if this method (and its segmentations_t equivalent) is still needed
    def Compartment(self, compartment: compartment_id_t, /) -> array_t:
        """"""
        if compartment is compartment_id_t.CELL:
            return self.cells_map

        if compartment is compartment_id_t.CYTOPLASM:
            return self.cytoplasms_map

        if compartment is compartment_id_t.NUCLEUS:
            return self.nuclei_map

        raise ValueError(f"{compartment}: Invalid compartment, or compartment is None")

    def BuildCellsFromMaps(self) -> None:
        """
        Segmentation are supposed to be binary (as opposed to already labeled)
        """
        labeled_cells, n_cells = mrph.label(
            self.cells_map, return_num=True, connectivity=1
        )

        if self.has_nuclei:
            labeled_nuclei = mrph.label(self.nuclei_map, connectivity=1)
            # TODO: code duplication with task>segmentation>_CorrectNumberOfNucleiPerCell
            for label in range(1, n_cells + 1):
                current_cell = labeled_cells == label
                inside_cell = labeled_nuclei[current_cell]
                labels_inside = nmpy.unique(inside_cell)
                assert labels_inside[0] == 0
                assert labels_inside.size > 1
                current_nuclei = labeled_nuclei == labels_inside[1]
                if labels_inside.size > 2:
                    for nucleus_lbl in labels_inside[2:]:
                        current_nuclei = nmpy.logical_or(
                            current_nuclei, labeled_nuclei == nucleus_lbl
                        )
                cell = cell_t.NewFromMaps(
                    cell_map=current_cell, nucleus_map=current_nuclei
                )
                self.cells.append(cell)
        else:
            for label in range(1, n_cells + 1):
                cell = cell_t.NewFromMaps(cell_map=labeled_cells == label)
                self.cells.append(cell)

    @property
    def has_cells(self) -> bool:
        """"""
        return (self.cells is not None) and (self.cells.__len__() > 0)

    # @property
    # def available_cell_features(self) -> Sequence[str]:
    #     """"""
    #     return self.cells[0].available_features
    #
    # @property
    # def available_cytoplasm_features(self) -> Sequence[str]:
    #     """"""
    #     if self.has_nuclei:
    #         return self.cells[0].cytoplasm.available_features
    #
    #     return ()
    #
    # @property
    # def available_nucleus_features(self) -> Sequence[str]:
    #     """"""
    #     if self.has_nuclei:
    #         return self.cells[0].nuclei[0].available_features
    #
    #     return ()

    @property
    def cell_areas(self) -> Sequence[int]:
        """"""
        labeled, n_cells = mrph.label(self.cells_map, return_num=True, connectivity=1)
        output = tuple(
            nmpy.count_nonzero(labeled == _lbl) for _lbl in range(1, n_cells + 1)
        )

        return output

    def CellAreaHistogram(
        self,
        /,
        *,
        n_bins: int = None,
        should_return_centers: bool = False,
        should_round_centers: bool = True,
    ) -> Tuple[array_t, array_t]:
        """"""
        areas = self.cell_areas

        if n_bins is None:
            n_bins = int(round(nmpy.sqrt(areas.__len__())))
        counts, edges = nmpy.histogram(areas, bins=n_bins)
        if should_return_centers:
            centers = 0.5 * (edges[:-1] + edges[1:])
            if should_round_centers:
                centers = nmpy.around(centers).astype(nmpy.uint64)
            edges = centers

        return counts, edges



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import functools as fctl
from itertools import starmap, zip_longest
from multiprocessing import Pool as pool_t
from typing import Any, Iterator, List, Sequence, Tuple, Union

import numpy as nmpy
import skimage.morphology as mrph

import cell_tracking_BC.in_out.text.progress as prgs
from cell_tracking_BC.task.feature import feature_computation_h
from cell_tracking_BC.type.compartment.base import compartment_id_t
from cell_tracking_BC.type.compartment.cell import cell_t
from cell_tracking_BC.type.compartment.cytoplasm import cytoplasm_t
from cell_tracking_BC.type.compartment.nucleus import nucleus_t
from cell_tracking_BC.type.segmentation.frame import segmentation_t


array_t = nmpy.ndarray


class segmentations_t(List[segmentation_t]):
    @classmethod
    def NewFromCellsMaps(
        cls,
        cells_maps: Sequence[array_t],
        /,
        *,
        nuclei_maps: Sequence[array_t] = None,
    ) -> segmentations_t:
        """"""
        instance = cls()

        if nuclei_maps is None:
            nuclei_maps = cells_maps.__len__() * [None]

        for cells_map, nuclei_map in zip(cells_maps, nuclei_maps):
            segmentation = segmentation_t.NewFromCellsMap(
                cells_map, nuclei_map=nuclei_map
            )
            instance.append(segmentation)

        return instance

    @property
    def length(self) -> int:
        """"""
        return self.__len__()

    def Compartments(self, compartment: compartment_id_t, /) -> Sequence[array_t]:
        """"""
        return tuple(_sgm.Compartment(compartment) for _sgm in self)

    def BuildCellsFromMaps(self) -> None:
        """
        Segmentation are supposed to be binary (as opposed to already labeled)
        """
        for segmentation in self:
            segmentation.BuildCellsFromMaps()

    def NCells(
        self, /, *, in_frame: Union[int, Sequence[int]] = None
    ) -> Union[int, Sequence[int]]:
        """
        in_frame: None=>total over the sequence
        """
        if in_frame is None:
            return sum(_cll.__len__() for _cll in self.cells_iterator)

        just_one = isinstance(in_frame, int)
        if self.has_cells:
            if just_one:
                in_frame = (in_frame,)

            output = in_frame.__len__() * [0]

            for f_idx, cells in enumerate(self.cells_iterator):
                if f_idx in in_frame:
                    output[in_frame.index(f_idx)] = cells.__len__()

            if just_one:
                output = output[0]
        elif just_one:
            output = 0
        else:
            output = in_frame.__len__() * (0,)

        return output

    @property
    def cells_iterator(self) -> Iterator[Sequence[cell_t]]:
        """"""
        for segmentation in self:
            yield segmentation.cells

    @property
    def cytoplasms_iterator(self) -> Iterator[Sequence[cytoplasm_t]]:
        """"""
        for segmentation in self:
            yield tuple(_cll.cytoplasm for _cll in segmentation.cells)

    @property
    def nuclei_iterator(self) -> Iterator[Sequence[Tuple[nucleus_t,...]]]:
        """"""
        for segmentation in self:
            nuclei = []
            for cell in segmentation.cells:
                nuclei.extend(cell.nuclei)
            yield tuple(nuclei)

    def AddCompartmentFeature(
        self,
        compartment: compartment_id_t,
        name: Union[str, Sequence[str]],
        Feature: feature_computation_h,
        /,
        frames: Union[Sequence[array_t], Sequence[Sequence[array_t]]] = None,
        should_run_in_parallel: bool = True,
        should_run_silently: bool = False,
        **kwargs,
    ) -> None:
        """
        name: If an str, then the value returned by Feature will be considered as a whole, whether it is actually a
        single value or a value container. If a sequence of str's, then the object returned by Feature will be iterated
        over, each element being matched with the corresponding name in "name".
        frames: if None, then geometrical feature, else radiometric feature.

        /!\ There is no already-existing check
        """
        if isinstance(name, str):
            description = f'Feature "{name}"'
        elif name.__len__() > 2:
            description = f'Feature "{name[0]}, ..., {name[-1]}"'
        else:
            description = f'Feature "{name[0]}, {name[1]}"'
        PreConfiguredFeatureFct = fctl.partial(Feature, **kwargs)

        with prgs.ProgressDesign(should_be_silent=should_run_silently) as progress:
            if compartment is compartment_id_t.CELL:
                compartment_iterator = self.cells_iterator
            elif compartment is compartment_id_t.CYTOPLASM:
                compartment_iterator = self.cytoplasms_iterator
            else:
                compartment_iterator = self.nuclei_iterator
            if frames is None:
                iterator = compartment_iterator
            else:
                iterator = zip(compartment_iterator, frames)
            progress_context = prgs.progress_context_t(
                progress,
                iterator,
                total=self.__len__(),
                description=description,
            )

            if should_run_in_parallel:
                pool = pool_t()
                MapFunctionOnList = pool.map
                StarMapFunctionOnList = pool.starmap
            else:
                pool = None
                MapFunctionOnList = map
                StarMapFunctionOnList = starmap

            if frames is None:
                if isinstance(name, str):
                    for compartments in progress_context.elements:
                        features = MapFunctionOnList(PreConfiguredFeatureFct, compartments)
                        for compartment, feature in zip(compartments, features):
                            compartment.AddFeature(name, feature)
                else:
                    names = name
                    for compartments in progress_context.elements:
                        multi_features = MapFunctionOnList(
                            PreConfiguredFeatureFct, compartments
                        )
                        for compartment, features in zip(compartments, multi_features):
                            for name, feature in zip(names, features):
                                compartment.AddFeature(name, feature)
            else:
                if isinstance(name, str):
                    for compartments, frame in progress_context.elements:
                        features = StarMapFunctionOnList(
                            PreConfiguredFeatureFct,
                            zip_longest(compartments, (frame,), fillvalue=frame),
                        )
                        for compartment, feature in zip(compartments, features):
                            compartment.AddFeature(name, feature)
                else:
                    names = name
                    for compartments, frame in progress_context.elements:
                        multi_features = StarMapFunctionOnList(
                            PreConfiguredFeatureFct,
                            zip_longest(compartments, (frame,), fillvalue=frame),
                        )
                        for compartment, features in zip(compartments, multi_features):
                            for name, feature in zip(names, features):
                                compartment.AddFeature(name, feature)

            if should_run_in_parallel:
                pool.close()
                pool.terminate()

    @property
    def cell_areas(self) -> Sequence[int]:
        """"""
        output = []

        for segmentation in self:
            labeled, n_cells = mrph.label(
                segmentation.cells_map, return_num=True, connectivity=1
            )
            areas = (
                nmpy.count_nonzero(labeled == _lbl) for _lbl in range(1, n_cells + 1)
            )
            output.extend(areas)

        return output

    def CellAreaHistogram(
        self,
        /,
        *,
        n_bins: int = None,
        should_return_centers: bool = False,
        should_round_centers: bool = True,
    ) -> Tuple[array_t, array_t]:
        """"""
        areas = self.cell_areas

        if n_bins is None:
            n_bins = int(round(nmpy.sqrt(areas.__len__())))
        counts, edges = nmpy.histogram(areas, bins=n_bins)
        if should_return_centers:
            centers = 0.5 * (edges[:-1] + edges[1:])
            if should_round_centers:
                centers = nmpy.around(centers).astype(nmpy.uint64)
            edges = centers

        return counts, edges

    def AsArray(
        self, /, *, compartment: Union[compartment_id_t, Sequence[compartment_id_t]] = None
    ) -> array_t:
        """"""
        if compartment is None:
            if self.has_nuclei:
                compartment = (
                    compartment_id_t.CELL,
                    compartment_id_t.CYTOPLASM,
                    compartment_id_t.NUCLEUS,
                )
            else:
                compartment = compartment_id_t.CELL

        if isinstance(compartment, compartment_id_t):
            return nmpy.dstack(self.Compartments(compartment))

        compartments = tuple(self.Compartments(_cpt) for _cpt in compartment)
        frames = (nmpy.dstack(_cpt) for _cpt in zip(*compartments))

        return nmpy.stack(tuple(frames), axis=3)

    def __getattribute__(self, name) -> Any:
        """"""
        try:
            output = object.__getattribute__(self, name)
        except AttributeError:
            if self.length > 0:
                output = object.__getattribute__(self[0], name)
            else:
                raise AttributeError(
                    f"{name}: Invalid attribute or attribute access on empty segmentations"
                )

        return output


segmentations_h = Union[Sequence[array_t], Sequence[segmentation_t], segmentations_t]


# def AllSegmentations(
#     segmentations: Union[Sequence[array_t], segmentations_t], /
# ) -> Tuple[all_versions_h, str]:
#     """"""
#     if isinstance(segmentations, segmentations_t):
#         compartments = segmentations[0].available_compartments
#         for compartment in compartments:
#                 output = segmentations.Compartments(
#                     compartment
#                 )
#     else:
#         output = segmentations
#
#     return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as nmpy
import scipy.interpolate as trpl
import skimage as skim


array_t = nmpy.ndarray


_LOCAL_HISTOGRAM_FILTER_WIDTH = 3.0
_RESCALING_FILTER_WIDTH = 9.0


def WithBackgroundSubtracted(
    image: array_t,
    segmentation: array_t,
    /,
    *,
    mode: str = None,
    dilation: int = 5,
) -> array_t:
    """
    mode: None (=mean), mean, median, interpolated
    Using the image and its segmentation.
    """
    missing_map = skim.morphology.binary_dilation(
        segmentation, footprint=skim.morphology.disk(dilation)
    )
    filled_map = nmpy.logical_not(missing_map)

    if (mode is None) or (mode == "mean"):
        output = image - nmpy.mean(image[filled_map])
    elif mode == "median":
        output = image - nmpy.median(image[filled_map])
    elif mode == "interpolated":
        output = image - _PChipInterpolatedBackground(image, filled_map, missing_map)
    else:
        raise NotImplementedError(
            f"{mode}: Invalid or unimplemented background subtraction mode"
        )

    output[output < 0.0] = 0.0

    return output


def _PChipInterpolatedBackground(
    incomplete: array_t, filled_map: array_t, missing_map: array_t, /
) -> array_t:
    """
    Uses pchip interpolation to avoid negative values.
    No smoothing is performed on the estimated background.
    """
    # Since pchip works in 1-D, interpolation has to be done in 2 steps
    # Step 1: row-wise interpolation
    row_wise = nmpy.array(incomplete, dtype=nmpy.float64)

    for col in range(row_wise.shape[1]):
        old_rows = nmpy.nonzero(filled_map[:, col])[0]
        new_rows = nmpy.nonzero(missing_map[:, col])[0]

        row_wise[new_rows, col] = trpl.pchip_interpolate(
            old_rows, incomplete[old_rows, col], new_rows
        )

    # Step 2: column-wise interpolation
    col_wise = nmpy.array(incomplete, dtype=nmpy.float64)

    for row in range(col_wise.shape[0]):
        old_cols = nmpy.nonzero(filled_map[row, :])[0]
        new_cols = nmpy.nonzero(missing_map[row, :])[0]

        col_wise[row, new_cols] = trpl.pchip_interpolate(
            old_cols, incomplete[row, old_cols], new_cols
        )

    output = nmpy.array(incomplete, dtype=nmpy.float64)
    output[missing_map] = 0.5 * (row_wise[missing_map] + col_wise[missing_map])

    return output
